@@ -1,9 +1,10 @@
 #!/usr/bin/env python
 # -*- coding: utf8 -*-
-#
+# pylint:disable=too-many-lines
+
 # library for communicating with an isc dhcp server over the omapi protocol
 #
-# Copyright 2010-2015 Cygnus Networks GmbH
+# Copyright 2010-2017 Cygnus Networks GmbH
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -33,18 +34,9 @@
 # dictionary = entry* 0x00 0x00
 # entry = key (net16str) value (net32str)
 
-__author__ = "Helmut Grohne, Dr. Torge Szczepanek"
-__copyright__ = "Cygnus Networks GmbH"
-__license__ = "Apache-2.0"
-__version__ = "0.4"
-__maintainer__ = "Dr. Torge Szczepanek"
-__email__ = "debian@cygnusnetworks.de"
-
-
-__all__ = []
-
 import binascii
 import struct
+import hashlib
 import hmac
 import io
 import logging
@@ -56,6 +48,16 @@ try:
 except NameError:
 	basestring = str  # pylint:disable=W0622
 
+__author__ = "Helmut Grohne, Dr. Torge Szczepanek"
+__copyright__ = "Cygnus Networks GmbH"
+__license__ = "Apache-2.0"
+__version__ = "0.8"
+__maintainer__ = "Dr. Torge Szczepanek"
+__email__ = "debian@cygnusnetworks.de"
+
+
+__all__ = []
+
 logger = logging.getLogger("pypureomapi")
 sysrand = random.SystemRandom()
 
@@ -77,12 +79,14 @@ def repr_opcode(opcode):
 	opmap = {1: "open", 2: "refresh", 3: "update", 4: "notify", 5: "status", 6: "delete"}
 	return opmap.get(opcode, "unknown (%d)" % opcode)
 
+
 __all__.append("OmapiError")
 
 
 class OmapiError(Exception):
 	"""OMAPI exception base class."""
 
+
 __all__.append("OmapiSizeLimitError")
 
 
@@ -91,6 +95,7 @@ class OmapiSizeLimitError(OmapiError):
 	def __init__(self):
 		OmapiError.__init__(self, "Packet size limit reached.")
 
+
 __all__.append("OmapiErrorNotFound")
 
 
@@ -100,6 +105,15 @@ class OmapiErrorNotFound(OmapiError):
 		OmapiError.__init__(self, "not found")
 
 
+__all__.append("OmapiErrorAttributeNotFound")
+
+
+class OmapiErrorAttributeNotFound(OmapiErrorNotFound):
+	"""Attribute not found."""
+	def __init__(self):  # pylint:disable=super-init-not-called
+		OmapiError.__init__(self, "attribute not found")  # pylint:disable=non-parent-init-called
+
+
 class OutBuffer(object):
 	"""Helper class for constructing network packets."""
 	sizelimit = 65536
@@ -352,12 +366,13 @@ class OmapiHMACMD5Authenticator(OmapiAuthenticatorBase):
 		@rtype: bytes
 		@returns: a signature of length self.authlen
 		"""
-		return hmac.HMAC(self.key, message).digest()
+		return hmac.HMAC(self.key, message, digestmod=hashlib.md5).digest()
+
 
 __all__.append("OmapiMessage")
 
 
-class OmapiMessage(object):
+class OmapiMessage(object):  # pylint:disable=too-many-instance-attributes
 	"""
 	@type authid: int
 	@ivar authid: The id of the message authenticator.
@@ -378,7 +393,7 @@ class OmapiMessage(object):
 	@ivar signature: A signature on this message as generated by an
 			authenticator.
 	"""
-	def __init__(self, authid=0, opcode=0, handle=0, tid=0, rid=0, message=None, obj=None, signature=b""):
+	def __init__(self, authid=0, opcode=0, handle=0, tid=0, rid=0, message=None, obj=None, signature=b""):  # pylint:disable=too-many-arguments
 		"""
 		Construct an OmapiMessage from the given fields. No error
 		checking is performed.
@@ -525,7 +540,7 @@ class OmapiMessage(object):
 		@rtype: str
 		@returns: a barely human readable representation in one line
 		"""
-		return ("authid=%d authlen=%d opcode=%s handle=%d tid=%d rid=%d message=%r obj=%r signature=%r") % (self.authid, len(self.signature), repr_opcode(self.opcode), self.handle, self.tid, self.rid, self.message, self.obj, self.signature)
+		return "authid=%d authlen=%d opcode=%s handle=%d tid=%d rid=%d message=%r obj=%r signature=%r" % (self.authid, len(self.signature), repr_opcode(self.opcode), self.handle, self.tid, self.rid, self.message, self.obj, self.signature)
 
 
 def parse_map(filterfun, parser):
@@ -559,7 +574,7 @@ def parse_chain(*args):
 	"""
 	items = []
 	for parser in args:
-		for element in parser(*items):
+		for element in parser(*items):  # pylint:disable=star-args
 			if element is None:
 				yield None
 			else:
@@ -672,7 +687,7 @@ class InBuffer(object):
 		>>> d = b"\\0\\0\\0\\x64\\0\\0\\0\\x18"
 		>>> next(InBuffer(d).parse_startup_message()).validate()
 		"""
-		return parse_map(lambda args: OmapiStartupMessage(*args), parse_chain(self.parse_net32int, lambda _: self.parse_net32int()))
+		return parse_map(lambda args: OmapiStartupMessage(*args), parse_chain(self.parse_net32int, lambda _: self.parse_net32int()))  # pylint:disable=star-args
 
 	def parse_message(self):
 		"""results in an OmapiMessage"""
@@ -686,7 +701,8 @@ class InBuffer(object):
 							lambda *_: self.parse_bindict(),  # object
 							lambda *args: self.parse_fixedbuffer(args[1]))  # signature
 		return parse_map(lambda args:  # skip authlen in args:
-				OmapiMessage(*(args[0:1] + args[2:])), parser)
+				OmapiMessage(*(args[0:1] + args[2:])), parser)  # pylint:disable=star-args
+
 
 if isinstance(bytes(b"x")[0], int):
 	def bytes_to_int_seq(b):
@@ -699,6 +715,7 @@ else:
 	def int_seq_to_bytes(s):
 		return "".join([chr(x) for x in s])  # raises ValueError
 
+
 __all__.append("pack_ip")
 
 
@@ -725,6 +742,7 @@ def pack_ip(ipstr):
 	parts = [int(x) for x in parts]  # raises ValueError
 	return int_seq_to_bytes(parts)  # raises ValueError
 
+
 __all__.append("unpack_ip")
 
 
@@ -749,6 +767,7 @@ def unpack_ip(fourbytes):
 		raise ValueError("given buffer is not exactly four bytes long")
 	return ".".join([str(x) for x in bytes_to_int_seq(fourbytes)])
 
+
 __all__.append("pack_mac")
 
 
@@ -776,6 +795,7 @@ def pack_mac(macstr):
 	parts = [int(part, 16) for part in parts]  # raises ValueError
 	return int_seq_to_bytes(parts)  # raises ValueError
 
+
 __all__.append("unpack_mac")
 
 
@@ -801,10 +821,9 @@ def unpack_mac(sixbytes):
 	return ":".join(["%2.2x".__mod__(x) for x in bytes_to_int_seq(sixbytes)])
 
 
-class LazyStr(object):
-
-	def __init__(self, function):
-		self.function = function
+class LazyStr(object):  # pylint:disable=too-few-public-methods
+	def __init__(self, fnc):
+		self.function = fnc
 
 	def __str__(self):
 		return self.function()
@@ -812,9 +831,10 @@ class LazyStr(object):
 
 class TCPClientTransport(object):
 	"""PEP 3156 dummy transport class to support OmapiProtocol class."""
-	def __init__(self, protocol, host, port):
+	def __init__(self, protocol, host, port, timeout=None):
 		self.protocol = protocol
 		self.connection = socket.socket()
+		self.connection.settimeout(timeout)
 		self.connection.connect((host, port))
 		self.protocol.connection_made(self)
 
@@ -925,11 +945,12 @@ class OmapiProtocol(object):
 		logger.debug("sending %s", LazyStr(message.dump_oneline))
 		self.transport.write(message.as_string())
 
+
 __all__.append("Omapi")
 
 
-class Omapi(object):
-	def __init__(self, hostname, port, username=None, key=None):
+class Omapi(object):  # pylint:disable=too-many-public-methods
+	def __init__(self, hostname, port, username=None, key=None, timeout=None):  # pylint:disable=too-many-arguments
 		"""
 		@type hostname: str
 		@type port: int
@@ -951,7 +972,7 @@ class Omapi(object):
 		if username is not None and key is not None:
 			newauth = OmapiHMACMD5Authenticator(username, key)
 
-		self.transport = TCPClientTransport(self.protocol, hostname, port)
+		self.transport = TCPClientTransport(self.protocol, hostname, port, timeout=timeout)
 
 		self.recv_protocol_initialization()
 
@@ -1051,6 +1072,169 @@ class Omapi(object):
 		self.protocol.defauth = authid
 		logger.debug("successfully initialized default authid %d", authid)
 
+	def lookup_ip_host(self, mac):
+		"""Lookup a host object with with given mac address.
+
+		@type mac: str
+		@raises ValueError:
+		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no lease object with the given mac could be found
+		@raises OmapiErrorAttributeNotFound: if lease could be found, but objects lacks a ip
+		@raises socket.error:
+		"""
+		res = self.lookup_by_host(mac=mac)
+		try:
+			return res["ip-address"]
+		except KeyError:
+			raise OmapiErrorAttributeNotFound()
+
+	def lookup_ip(self, mac):
+		"""Look for a lease object with given mac address and return the
+		assigned ip address.
+
+		@type mac: str
+		@rtype: str or None
+		@raises ValueError:
+		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no lease object with the given mac could be found
+		@raises OmapiErrorAttributeNotFound: if lease could be found, but objects lacks a ip
+		@raises socket.error:
+		"""
+		res = self.lookup_by_lease(mac=mac)
+		try:
+			return res["ip-address"]
+		except KeyError:
+			raise OmapiErrorAttributeNotFound()
+
+	def lookup_mac(self, ip):
+		"""Look up a lease object with given ip address and return the
+		associated mac address.
+
+		@type ip: str
+		@rtype: str or None
+		@raises ValueError:
+		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no lease object with the given ip could be found
+		@raises OmapiErrorAttributeNotFound: if lease could be found, but objects lacks a mac
+		@raises socket.error:
+		"""
+		res = self.lookup_by_lease(ip=ip)
+		try:
+			return res["hardware-address"]
+		except KeyError:
+			raise OmapiErrorAttributeNotFound()
+
+	def lookup_host(self, name):
+		"""Look for a host object with given name and return the
+		name, mac, and ip address
+
+		@type name: str
+		@rtype: dict or None
+		@raises ValueError:
+		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no host object with the given name could be found
+		@raises OmapiErrorAttributeNotFound: if lease could be found, but objects lacks ip, mac or name
+		@raises socket.error:
+		"""
+		res = self.lookup_by_host(name=name)
+		try:
+			return dict(ip=res["ip-address"], mac=res["hardware-address"], hostname=res["name"].decode('utf-8'))
+		except KeyError:
+			raise OmapiErrorAttributeNotFound()
+
+	def lookup_host_host(self, mac):
+		"""Look for a host object with given mac address and return the
+		name, mac, and ip address
+
+		@type mac: str
+		@rtype: dict or None
+		@raises ValueError:
+		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no host object with the given mac address could be found
+		@raises OmapiErrorAttributeNotFound: if lease could be found, but objects lacks ip, mac or name
+		@raises socket.error:
+		"""
+		res = self.lookup_by_host(mac=mac)
+		try:
+			return dict(ip=res["ip-address"], mac=res["hardware-address"], name=res["name"].decode('utf-8'))
+		except KeyError:
+			raise OmapiErrorAttributeNotFound()
+
+	def lookup_hostname(self, ip):
+		"""Look up a lease object with given ip address and return the associated client hostname.
+
+		@type ip: str
+		@rtype: str or None
+		@raises ValueError:
+		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no lease object with the given ip address could be found
+		@raises OmapiErrorAttributeNotFound: if lease could be found, but objects lacks a hostname
+		@raises socket.error:
+		"""
+		res = self.lookup_by_lease(ip=ip)
+		if "client-hostname" not in res:
+			raise OmapiErrorAttributeNotFound()
+		return res["client-hostname"].decode('utf-8')
+
+	def lookup_by_host(self, **kwargs):
+		return self.__lookup("host", **kwargs)
+
+	def lookup_by_lease(self, **kwargs):
+		return self.__lookup("lease", **kwargs)
+
+	def __lookup(self, ltype, **kwargs):
+		"""Generic Lookup function
+
+		@type ltype: str
+		@type rvalues: list
+		@type ip: str
+		@type mac: str
+		@type name: str
+		@rtype: dict or str (if len(rvalues) == 1) or None
+		@raises ValueError:
+		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no host object with the given name
+				could be found or the object lacks an ip address or mac
+		@raises socket.error:
+		"""
+		ltype_utf = ltype.encode("utf-8")
+		assert ltype_utf in [b"host", b"lease"]
+		msg = OmapiMessage.open(ltype_utf)
+		for k in kwargs:
+			if k == "raw":
+				continue
+			_k = k.replace("_", "-")
+			if _k in ["ip", "ip-address"]:
+				msg.obj.append((b"ip-address", pack_ip(kwargs[k])))
+			elif _k in ["mac", "hardware-address"]:
+				msg.obj.append((b"hardware-address", pack_mac(kwargs[k])))
+				msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
+			elif _k == "name":
+				msg.obj.append((b"name", kwargs[k].encode('utf-8')))
+			else:
+				msg.obj.append((str(k).encode(), kwargs[k].encode('utf-8')))
+		response = self.query_server(msg)
+		if response.opcode != OMAPI_OP_UPDATE:
+			raise OmapiErrorNotFound()
+		if "raw" in kwargs and kwargs["raw"]:
+			return dict(response.obj)
+		res = dict()
+		for k, v in dict(response.obj).items():
+			_k = k.decode('utf-8')
+			try:
+				if _k == "ip-address":
+					v = unpack_ip(v)
+				elif _k in ["hardware-address"]:
+					v = unpack_mac(v)
+				elif _k in ["starts", "ends", "tstp", "tsfp", "atsfp", "cltt", "subnet", "pool", "state", "hardware-type"]:
+					v = struct.unpack(">I", v)[0]
+				elif _k in ["flags"]:
+					v = struct.unpack(">I", v)[0]
+			except struct.error:
+				pass
+			res[_k] = v
+		return res
+
 	def add_host(self, ip, mac):
 		"""Create a host object with given ip address and and mac address.
 
@@ -1070,12 +1254,87 @@ class Omapi(object):
 		if response.opcode != OMAPI_OP_UPDATE:
 			raise OmapiError("add failed")
 
+	def add_host_supersede_name(self, ip, mac, name):  # pylint:disable=E0213
+		"""Add a host with a fixed-address and override its hostname with the given name.
+		@type self: Omapi
+		@type ip: str
+		@type mac: str
+		@type name: str
+		@raises ValueError:
+		@raises OmapiError:
+		@raises socket.error:
+		"""
+		msg = OmapiMessage.open(b"host")
+		msg.message.append((b"create", struct.pack("!I", 1)))
+		msg.message.append((b"exclusive", struct.pack("!I", 1)))
+		msg.obj.append((b"hardware-address", pack_mac(mac)))
+		msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
+		msg.obj.append((b"ip-address", pack_ip(ip)))
+		msg.obj.append((b"name", name.encode('utf-8')))
+		msg.obj.append((b"statements", 'supersede host-name "{0}";'.format(name).encode('utf-8')))
+		response = self.query_server(msg)
+		if response.opcode != OMAPI_OP_UPDATE:
+			raise OmapiError("add failed")
+
+	def add_host_without_ip(self, mac):
+		"""Create a host object with given mac address without assigning a static ip address.
+		@type mac: str
+		@raises ValueError:
+		@raises OmapiError:
+		@raises socket.error:
+		"""
+		msg = OmapiMessage.open(b"host")
+		msg.message.append((b"create", struct.pack("!I", 1)))
+		msg.message.append((b"exclusive", struct.pack("!I", 1)))
+		msg.obj.append((b"hardware-address", pack_mac(mac)))
+		msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
+		response = self.query_server(msg)
+		if response.opcode != OMAPI_OP_UPDATE:
+			raise OmapiError("add failed")
+
+	def add_host_supersede(self, ip, mac, name, hostname=None, router=None, domain=None):  # pylint:disable=too-many-arguments
+		"""Create a host object with given ip, mac, name, hostname, router and
+		domain. hostname, router and domain are optional arguments.
+
+		@type ip: str
+		@type mac: str
+		@type name: str
+		@type hostname: str
+		@type router: str
+		@type domain: str
+		@raises OmapiError:
+		@raises socket.error:
+		"""
+		stmts = []
+
+		msg = OmapiMessage.open(b"host")
+		msg.message.append((b"create", struct.pack("!I", 1)))
+		msg.obj.append((b"name", name))
+		msg.obj.append((b"hardware-address", pack_mac(mac)))
+		msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
+		msg.obj.append((b"ip-address", pack_ip(ip)))
+		if hostname:
+			stmts.append('supersede host-name "{0}";\n '.format(hostname))
+		if router:
+			stmts.append('supersede routers {0};\n '.format(router))
+		if domain:
+			stmts.append('supersede domain-name "{0}";'.format(domain))
+		if stmts:
+			encoded_stmts = "".join(stmts).encode("utf-8")
+			msg.obj.append((b"statements", encoded_stmts))
+
+		response = self.query_server(msg)
+		if response.opcode != OMAPI_OP_UPDATE:
+			raise OmapiError("add failed")
+
 	def del_host(self, mac):
 		"""Delete a host object with with given mac address.
 
 		@type mac: str
 		@raises ValueError:
 		@raises OmapiError:
+		@raises OmapiErrorNotFound: if no lease object with the given
+				mac address could be found
 		@raises socket.error:
 		"""
 		msg = OmapiMessage.open(b"host")
@@ -1090,72 +1349,55 @@ class Omapi(object):
 		if response.opcode != OMAPI_OP_STATUS:
 			raise OmapiError("delete failed")
 
-	def lookup_ip(self, mac):
-		"""Look for a lease object with given mac address and return the
-		assigned ip address.
-
-		@type mac: str
-		@rtype: str or None
-		@raises ValueError:
-		@raises OmapiError:
-		@raises OmapiErrorNotFound: if no lease object with the given mac
-				address could be found or the object lacks an ip address
-		@raises socket.error:
+	def add_group(self, groupname, statements):
 		"""
-		msg = OmapiMessage.open(b"lease")
-		msg.obj.append((b"hardware-address", pack_mac(mac)))
+		Adds a group
+		@type groupname: bytes
+		@type statements: str
+		"""
+		msg = OmapiMessage.open(b"group")
+		msg.message.append(("create", struct.pack("!I", 1)))
+		msg.obj.append(("name", groupname))
+		msg.obj.append(("statements", statements))
 		response = self.query_server(msg)
 		if response.opcode != OMAPI_OP_UPDATE:
-			raise OmapiErrorNotFound()
-		try:
-			return unpack_ip(dict(response.obj)[b"ip-address"])
-		except KeyError:  # ip-address
-			raise OmapiErrorNotFound()
+			raise OmapiError("add group failed")
 
-	def lookup_mac(self, ip):
-		"""Look up a lease object with given ip address and return the
-		associated mac address.
-
-		@type ip: str
-		@rtype: str or None
-		@raises ValueError:
-		@raises OmapiError:
-		@raises OmapiErrorNotFound: if no lease object with the given ip
-				address could be found or the object lacks a mac address
-		@raises socket.error:
+	def add_host_with_group(self, ip, mac, groupname):
 		"""
-		msg = OmapiMessage.open(b"lease")
-		msg.obj.append((b"ip-address", pack_ip(ip)))
-		response = self.query_server(msg)
-		if response.opcode != OMAPI_OP_UPDATE:
-			raise OmapiErrorNotFound()
-		try:
-			return unpack_mac(dict(response.obj)[b"hardware-address"])
-		except KeyError:  # hardware-address
-			raise OmapiErrorNotFound()
-
-	def add_host_supersede_name(omapi, ip, mac, name):  # pylint:disable=E0213
-		"""Add a host with a fixed-address and override its hostname with the given name.
-		@type omapi: Omapi
+		Adds a host with given ip and mac in a group named groupname
 		@type ip: str
 		@type mac: str
-		@type name: str
-		@raises ValueError:
-		@raises OmapiError:
-		@raises socket.error:
+		@type groupname: str
 		"""
 		msg = OmapiMessage.open(b"host")
-		msg.message.append((b"create", struct.pack("!I", 1)))
-		msg.message.append((b"exclusive", struct.pack("!I", 1)))
-		msg.obj.append((b"hardware-address", pack_mac(mac)))
-		msg.obj.append((b"hardware-type", struct.pack("!I", 1)))
-		msg.obj.append((b"ip-address", pack_ip(ip)))
-		msg.obj.append((b"name", name))
-		msg.obj.append((b"statements", str.encode('supersede host-name "%s";' % name)))
-		response = omapi.query_server(msg)
+		msg.message.append(("create", struct.pack("!I", 1)))
+		msg.message.append(("exclusive", struct.pack("!I", 1)))
+		msg.obj.append(("hardware-address", pack_mac(mac)))
+		msg.obj.append(("hardware-type", struct.pack("!I", 1)))
+		msg.obj.append(("ip-address", pack_ip(ip)))
+		msg.obj.append(("group", groupname))
+		response = self.query_server(msg)
 		if response.opcode != OMAPI_OP_UPDATE:
 			raise OmapiError("add failed")
 
-if __name__ == '__main__':
+	def change_group(self, name, group):
+		"""Change the group of a host given the name of the host.
+		@type name: str
+		@type group: str
+		"""
+		m1 = OmapiMessage.open(b"host")
+		m1.update_object(dict(name=name))
+		r1 = self.query_server(m1)
+		if r1.opcode != OMAPI_OP_UPDATE:
+			raise OmapiError("opening host %s failed" % name)
+		m2 = OmapiMessage.update(r1.handle)
+		m2.update_object(dict(group=group))
+		r2 = self.query_server(m2)
+		if r2.opcode != OMAPI_OP_UPDATE:
+			raise OmapiError("changing group of host %s to %s failed" % (name, group))
+
+
+if __name__ == "__main__":
 	import doctest
 	doctest.testmod()
